import unittest
from random import randint
from kyber_py.modules.modules import ModuleKyber


class TestModuleKyber(unittest.TestCase):
    M = ModuleKyber()
    R = M.ring

    def test_decode_vector(self):
        for _ in range(100):
            k = randint(1, 5)
            v = self.M.random_element(k, 1)
            v_bytes = v.encode(12)
            self.assertEqual(v, self.M.decode_vector(v_bytes, k, 12))

    def test_recode_vector_wrong_length(self):
        self.assertRaises(
            ValueError, lambda: self.M.decode_vector(b"1", 2, 12)
        )



from kyber_py.ml_kem import ML_KEM_512, ML_KEM_768, ML_KEM_1024
import cProfile
from time import time


def profile_ml_kem(ML_KEM):
    (ek, dk) = ML_KEM.keygen()
    (K, c) = ML_KEM.encaps(ek)

    gvars = {}
    lvars = {"ML_KEM": ML_KEM, "c": c, "ek": ek, "dk": dk}

    cProfile.runctx(
        "[ML_KEM.keygen() for _ in range(100)]",
        globals=gvars,
        locals=lvars,
        sort=1,
    )
    cProfile.runctx(
        "[ML_KEM.encaps(ek) for _ in range(100)]",
        globals=gvars,
        locals=lvars,
        sort=1,
    )
    cProfile.runctx(
        "[ML_KEM.decaps(dk, c) for _ in range(100)]",
        globals=gvars,
        locals=lvars,
        sort=1,
    )


def benchmark_ml_kem(ML_KEM, name, count):
    keygen_times = []
    enc_times = []
    dec_times = []

    for _ in range(count):
        t0 = time()
        ek, dk = ML_KEM.keygen()
        keygen_times.append(time() - t0)

        t1 = time()
        _, c = ML_KEM.encaps(ek)
        enc_times.append(time() - t1)

        t2 = time()
        _ = ML_KEM.decaps(dk, c)
        dec_times.append(time() - t2)

    avg_keygen = sum(keygen_times) / count
    avg_enc = sum(enc_times) / count
    avg_dec = sum(dec_times) / count
    print(
        f" {name:11} |"
        f"{avg_keygen*1000:7.2f}ms | {1/avg_keygen:10.2f} |"
        f"{avg_enc*1000:6.2f}ms | {1/avg_enc:9.2f} |"
        f"{avg_dec*1000:6.2f}ms | {1/avg_dec:7.2f} |"
    )


if __name__ == "__main__":
    count = 1000
    # common banner
    print("-" * 80)
    print(
        "   Params    |  keygen  |  keygen/s  |  encap  |  encap/s  "
        "|  decap  |  decap/s"
    )
    print("-" * 80)
    benchmark_ml_kem(ML_KEM_512, "ML-KEM-512", count)
    benchmark_ml_kem(ML_KEM_768, "ML-KEM-768", count)
    benchmark_ml_kem(ML_KEM_1024, "ML-KEM-1024", count)



import unittest
from kyber_py.polynomials.polynomials import PolynomialRingKyber


class TestModuleKyber(unittest.TestCase):
    R = PolynomialRingKyber()

    def test_decode(self):
        for _ in range(10):
            f = self.R.random_element()
            f_bytes = f.encode(12)
            self.assertEqual(f, self.R.decode(f_bytes, 12))

    def test_decode_wrong_length(self):
        self.assertRaises(ValueError, lambda: self.R.decode(b"1", 12))

    def test_call(self):
        self.assertEqual(1, self.R(1))
        self.assertRaises(TypeError, lambda: self.R("a"))


class TestPolynomial(unittest.TestCase):
    R = PolynomialRingKyber()

    def test_ntt_transform(self):
        f = self.R.random_element()
        g_hat = self.R.random_element().to_ntt()

        self.assertEqual(f, f.to_ntt().from_ntt())
        self.assertEqual(g_hat, g_hat.from_ntt().to_ntt())

        self.assertRaises(TypeError, lambda: f.from_ntt())
        self.assertRaises(TypeError, lambda: g_hat.to_ntt())

    def test_add_failure(self):
        f1 = self.R.random_element()
        self.assertRaises(NotImplementedError, lambda: f1 + "a")

    def test_sub_failure(self):
        f1 = self.R.random_element()
        self.assertRaises(NotImplementedError, lambda: f1 - "a")

    def test_mul_failure(self):
        f1 = self.R.random_element()
        self.assertRaises(NotImplementedError, lambda: f1 * "a")

    def test_pow_failure(self):
        f1 = self.R.random_element()
        self.assertRaises(TypeError, lambda: f1 ** "a")

    def test_add_polynomials(self):
        zero = self.R(0)
        for _ in range(10):
            f1 = self.R.random_element()
            f2 = self.R.random_element()
            f3 = self.R.random_element()

            self.assertEqual(f1 + zero, f1)
            self.assertEqual(f1 + f2, f2 + f1)
            self.assertEqual(f1 + (f2 + f3), (f1 + f2) + f3)

            f2 = f1
            f2 += f1
            self.assertEqual(f1 + f1, f2)

    def test_sub_polynomials(self):
        zero = self.R(0)
        for _ in range(10):
            f1 = self.R.random_element()
            f2 = self.R.random_element()
            f3 = self.R.random_element()

            self.assertEqual(f1 - zero, f1)
            self.assertEqual(f3 - f3, zero)
            self.assertEqual(f3 - 0, f3)
            self.assertEqual(0 - f3, -f3)
            self.assertEqual(f1 - f2, -(f2 - f1))
            self.assertEqual(f1 - (f2 - f3), (f1 - f2) + f3)

            f2 = f1
            f2 -= f1
            self.assertEqual(f2, zero)

    def test_mul_polynomials(self):
        zero = self.R(0)
        one = self.R(1)
        for _ in range(10):
            f1 = self.R.random_element()
            f2 = self.R.random_element()
            f3 = self.R.random_element()

            self.assertEqual(f1 * zero, zero)
            self.assertEqual(f1 * one, f1)
            self.assertEqual(f1 * f2, f2 * f1)
            self.assertEqual(f1 * (f2 * f3), (f1 * f2) * f3)
            self.assertEqual(2 * f1, f1 + f1)
            self.assertEqual(2 * f1, f1 * 2)

            f2 = f1
            f2 *= f2
            self.assertEqual(f1 * f1, f2)

    def test_pow_polynomials(self):
        one = self.R(1)
        for _ in range(10):
            f1 = self.R.random_element()

            self.assertEqual(one, f1**0)
            self.assertEqual(f1, f1**1)
            self.assertEqual(f1 * f1, f1**2)
            self.assertEqual(f1 * f1 * f1, f1**3)
            self.assertRaises(ValueError, lambda: f1 ** (-1))

    def test_add_failure_ntt(self):
        f1 = self.R.random_element().to_ntt()
        self.assertRaises(NotImplementedError, lambda: f1 + "a")

    def test_sub_failure_ntt(self):
        f1 = self.R.random_element().to_ntt()
        self.assertRaises(NotImplementedError, lambda: f1 - "a")

    def test_mul_failure_ntt(self):
        f1 = self.R.random_element().to_ntt()
        self.assertRaises(NotImplementedError, lambda: f1 * "a")

    def test_pow_failure_ntt(self):
        f1 = self.R.random_element().to_ntt()
        self.assertRaises(TypeError, lambda: f1 ** "a")

    def test_add_polynomials_ntt(self):
        zero_hat = self.R(0).to_ntt()
        for _ in range(10):
            f1_hat = self.R.random_element().to_ntt()
            f2_hat = self.R.random_element().to_ntt()
            f3_hat = self.R.random_element().to_ntt()

            self.assertEqual(f1_hat + zero_hat, f1_hat)
            self.assertEqual(f1_hat + f2_hat, f2_hat + f1_hat)
            self.assertEqual(
                f1_hat + (f2_hat + f3_hat), (f1_hat + f2_hat) + f3_hat
            )

            f2_hat = f1_hat
            f2_hat += f1_hat
            self.assertEqual(f1_hat + f1_hat, f2_hat)

    def test_sub_polynomials_ntt(self):
        zero_hat = self.R(0).to_ntt()
        for _ in range(10):
            f1_hat = self.R.random_element().to_ntt()
            f2_hat = self.R.random_element().to_ntt()
            f3_hat = self.R.random_element().to_ntt()

            self.assertEqual(f1_hat - zero_hat, f1_hat)
            self.assertEqual(f3_hat - f3_hat, zero_hat)
            self.assertEqual(f3_hat - 0, f3_hat)
            self.assertEqual(0 - f3_hat, -f3_hat)
            self.assertEqual(
                f1_hat - (f2_hat - f3_hat), (f1_hat - f2_hat) + f3_hat
            )

            f2_hat = f1_hat
            f2_hat -= f1_hat
            self.assertEqual(f2_hat, zero_hat)

    def test_mul_polynomials_ntt(self):
        zero_hat = self.R(0).to_ntt()
        one_hat = self.R(1).to_ntt()
        for _ in range(10):
            f1_hat = self.R.random_element().to_ntt()
            f2_hat = self.R.random_element().to_ntt()
            f3_hat = self.R.random_element().to_ntt()

            self.assertEqual(f1_hat * zero_hat, zero_hat)
            self.assertEqual(f1_hat * one_hat, f1_hat)
            self.assertEqual(f1_hat * f2_hat, f2_hat * f1_hat)
            self.assertEqual(
                f1_hat * (f2_hat * f3_hat), (f1_hat * f2_hat) * f3_hat
            )
            self.assertEqual(2 * f1_hat, f1_hat + f1_hat)
            self.assertEqual(2 * f1_hat, f1_hat * 2)

            f2_hat = f1_hat
            f2_hat *= f2_hat
            self.assertEqual(f1_hat * f1_hat, f2_hat)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Code location -----------------------------------------------------------

import os
import sys

sys.path.insert(0, os.path.abspath("../../src"))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "kyber-py"
copyright = "2024, Giacomo Pope"
author = "Giacomo Pope"
release = "0.3.0-alpha1"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
intersphinx_mapping = {"Python": ("https://docs.python.org/", None)}

autodoc_default_options = {
    "undoc-members": True,
    "inherited-members": True,
}



import unittest
import os
import pytest
from kyber_py.kyber import Kyber512, Kyber768, Kyber1024
from kyber_py.drbg.aes256_ctr_drbg import AES256_CTR_DRBG


def parse_kat_data(data):
    parsed_data = {}
    count_blocks = data.split("\n\n")
    for block in count_blocks[:-1]:
        block_data = block.split("\n")
        count, seed, pk, sk, ct, ss = [
            line.split(" = ")[-1] for line in block_data
        ]
        parsed_data[int(count)] = {
            "seed": bytes.fromhex(seed),
            "pk": bytes.fromhex(pk),
            "sk": bytes.fromhex(sk),
            "ct": bytes.fromhex(ct),
            "ss": bytes.fromhex(ss),
        }
    return parsed_data


class TestKyber(unittest.TestCase):
    """
    Test Kyber levels for internal consistency by generating keypairs and
    shared secrets.
    """

    def generic_test_kyber(self, Kyber, count):
        for _ in range(count):
            pk, sk = Kyber.keygen()
            for _ in range(count):
                key, c = Kyber.encaps(pk)

                # Correct decaps works
                _key = Kyber.decaps(sk, c)
                self.assertEqual(key, _key)

                # Incorrect ct does not work
                _bad_ct = bytes([0] * len(c))
                _bad = Kyber.decaps(sk, _bad_ct)
                self.assertNotEqual(key, _bad)

    def test_kyber512(self):
        self.generic_test_kyber(Kyber512, 5)

    def test_kyber768(self):
        self.generic_test_kyber(Kyber768, 5)

    def test_kyber1024(self):
        self.generic_test_kyber(Kyber1024, 5)

    def test_xof_failure(self):
        self.assertRaises(ValueError, lambda: Kyber512._xof(b"1", b"2", b"3"))

    def test_prf_failure(self):
        self.assertRaises(ValueError, lambda: Kyber512._prf(b"1", b"2", 32))


class TestKyberDeterministic(unittest.TestCase):
    """
    Ensure that deterministic DRBG is deterministic!

    Uses AES256 CTR DRBG for randomness.
    Note: requires pycryptodome for AES impl.
    (Seemed overkill to code my own AES for Kyber)
    """

    def generic_test_kyber_deterministic(self, Kyber, count):
        """
        First we generate five pk,sk pairs
        from the same seed and make sure
        they're all the same
        """
        seed = os.urandom(48)
        pk_output = []
        for _ in range(count):
            Kyber.set_drbg_seed(seed)
            pk, sk = Kyber.keygen()
            pk_output.append(pk + sk)
        self.assertEqual(len(pk_output), 5)
        self.assertEqual(len(set(pk_output)), 1)

        """
        Now given a fixed keypair make sure
        that c,key are the same for a fixed seed
        """
        key_output = []
        seed = os.urandom(48)
        pk, sk = Kyber.keygen()
        for _ in range(count):
            Kyber.set_drbg_seed(seed)
            key, c = Kyber.encaps(pk)
            _key = Kyber.decaps(sk, c)
            # Check key derivation works
            self.assertEqual(key, _key)
            key_output.append(c + key)
        self.assertEqual(len(key_output), count)
        self.assertEqual(len(set(key_output)), 1)

    def test_kyber512_deterministic(self):
        self.generic_test_kyber_deterministic(Kyber512, 5)

    def test_kyber768_deterministic(self):
        self.generic_test_kyber_deterministic(Kyber768, 5)

    def test_kyber1024_deterministic(self):
        self.generic_test_kyber_deterministic(Kyber1024, 5)


def data_parse(filename):
    # Set DRBG to generate seeds
    entropy_input = bytes([i for i in range(48)])
    rng = AES256_CTR_DRBG(entropy_input)

    with open(filename) as f:
        kat_data = f.read()
        parsed_data = parse_kat_data(kat_data)

    return [(rng.random_bytes(48), i) for i in parsed_data.values()]


@pytest.mark.parametrize(
    "Kyber, seed, data",
    [
        (kem, seed, param)
        for kem, filename in [
            (Kyber512, "assets/PQCkemKAT_1632.rsp"),
            (Kyber768, "assets/PQCkemKAT_2400.rsp"),
            (Kyber1024, "assets/PQCkemKAT_3168.rsp"),
        ]
        for seed, param in data_parse(filename)
    ],
    ids=[
        f"{kem}-test-{num}"
        for kem in ["Kyber512", "Kyber768", "Kyber1024"]
        for num in range(100)
    ],
)
def test_generic_kyber_known_answer(Kyber, seed, data):

    # Set the seed and check it matches the KAT
    assert seed == data["seed"]

    # Seed DRBG with KAT seed
    Kyber.set_drbg_seed(seed)

    # Assert keygen matches
    pk, sk = Kyber.keygen()
    assert pk == data["pk"]
    assert sk == data["sk"]

    # Assert encapsulation matches
    ss, ct = Kyber.encaps(pk)
    assert ss == data["ss"]
    assert ct == data["ct"]

    # Assert decapsulation matches
    _ss = Kyber.decaps(sk, ct)
    assert _ss == data["ss"]



"""
The parameters defined in the FIPS 203 document.

Includes the ML-KEM-512, ML-KEM-768, and ML-KEM-1024 parameters
and initialised objects with them.
"""

from .ml_kem import ML_KEM

# TODO: we can only allow a user to select one of the following three
# we should maybe put these into the class and only allow a user to
# select 128, 192 or 256 bit security.
DEFAULT_PARAMETERS = {
    "ML512": {"k": 2, "eta_1": 3, "eta_2": 2, "du": 10, "dv": 4},
    "ML768": {"k": 3, "eta_1": 2, "eta_2": 2, "du": 10, "dv": 4},
    "ML1024": {"k": 4, "eta_1": 2, "eta_2": 2, "du": 11, "dv": 5},
}
"""Parameters for the :py:obj:`.ML_KEM` objects."""

ML_KEM_512 = ML_KEM(DEFAULT_PARAMETERS["ML512"])
"""
Key exchange object that uses ML-KEM-512 parameters internally.

Provides about 128 bit level of security.

Part of stable API.
"""

ML_KEM_768 = ML_KEM(DEFAULT_PARAMETERS["ML768"])
"""
Key exchange object that uses ML-KEM-768 parameters internally.

Provides about 192 bit level of security.

Part of stable API.
"""

ML_KEM_1024 = ML_KEM(DEFAULT_PARAMETERS["ML1024"])
"""
Key exchange object that uses ML-KEM-1024 parameters internally.

Provides about 256 bit level of security.

Part of stable API.
"""
